(require 'asdf)
(require 'pal)
(require 'split-sequence)
(require 'usocket)

(load "common")

(defpackage :blaster-client
  (:use :common-lisp :blaster-common)
  (:export #:blaster))

(in-package :blaster-client)

(defparameter *ground-image* nil)
(defparameter *wall-image* nil)
(defparameter *obstacle-image* nil)
(defparameter *player-image* nil)
(defparameter *dead-image* nil)
(defparameter *bomb-image* nil)
(defparameter *explosion-image* nil)

(defun load-images ()
  (setf *ground-image* (pal:load-image "Images/ground.png")
        *wall-image* (pal:load-image "Images/wall.png")
        *obstacle-image* (pal:load-image "Images/obstacle.png")
        *player-image* (pal:load-image "Images/player.png")
        *dead-image* (pal:load-image "Images/dead.png")
        *bomb-image* (pal:load-image "Images/bomb.png")
        *explosion-image* (pal:load-image "Images/explosion.png")))

(defparameter *explosion-sound* nil)

(defun load-sounds ()
  (handler-case
      (setf *explosion-sound* (pal:load-sample "Sounds/explosion.wav"))
    (error () (format t "Cannot open sound device, disabling sounds~%"))))

(defparameter +screen-width+ 800)
(defparameter +screen-height+ 600)

(defparameter +player-width+ 50)
(defparameter +player-height+ 100)

(defparameter +player-screen-x-offset+ 25)
(defparameter +player-screen-y-offset+ 85)

(defparameter +bomb-screen-y-offset+ 14)

(defun world->screen-coords (x y)
  (values (floor (* x 65) +cell-size+)
          (floor (* y 48) +cell-size+)))

(defvar *player-animation-step* 1)

(defgeneric draw-entity (entity map))

(defmethod draw-entity ((entity player) map)
  (multiple-value-bind (x y)
      (world->screen-coords (+ (floor +cell-size+ 2)
                               (* (x entity) +cell-size+)
                               (dx entity))
                            (+ (floor +cell-size+ 2)
                               (* (y entity) +cell-size+)
                               (dy entity)))
    (setf x (- x +player-screen-x-offset+)
          y (- y +player-screen-y-offset+))
    (case (status entity)
      (:playing
      (pal:with-clipping (x (- +screen-height+ (+ y +player-height+))
                             +player-width+ +player-height+)
         (let ((step (if (moving-p entity)
                         *player-animation-step*
                         0)))
           (pal:draw-image *player-image*
                           (pal:v (- x (* +player-width+ step))
                                  (- y (* +player-height+
                                          (ecase (direction entity)
                                            (:north 1)
                                            (:south 0)
                                            (:west 2)
                                            (:east 3)))))))))
      (:dead
       (pal:draw-image *dead-image* (pal:v x y))))))

(defmethod draw-entity ((entity bomb) map)
  (if (null (explosion-step entity))
      (multiple-value-bind (x y)
          (world->screen-coords (* (x entity) +cell-size+)
                                (* (y entity) +cell-size+))
        (decf y +bomb-screen-y-offset+)
        (pal:draw-image *bomb-image* (pal:v x y)))
      (loop for (explosion-x . explosion-y)
            in (bomb-kill-zone entity map)
            do (multiple-value-bind (x y)
                   (world->screen-coords (* explosion-x +cell-size+)
                                         (* explosion-y +cell-size+))
                 (decf y +bomb-screen-y-offset+)
                 (pal:with-clipping (x (- +screen-height+ (+ y 48)) 65 48)
                   (pal:draw-image *explosion-image*
                                   (pal:v (- x (* 65 (explosion-step entity)))
                                          y)))))))

(defun draw-game (map blocks players bombs)
  (pal:clear-screen 40 40 40)
  (loop for j below (height map)
        for y = 0 then (+ y 48)
        do (loop for i below (width map)
                 for x = 0 then (+ x 65)
                 do (pal:draw-image (case (aref (data map) j i)
                                      ((nil) *ground-image*)
                                      (:wall *wall-image*)
                                      (:obstacle *obstacle-image*))
                                    (pal:v x y))))
  (loop with entities = (append (copy-list players) (copy-list bombs))
        for entity in (sort entities #'< :key #'y)
        do (draw-entity entity map))
  (incf *player-animation-step*)
  (when (> *player-animation-step* 4)
    (setf *player-animation-step* 1)))

(defun parse-response (line)
  (let ((trimmed (and line (string-trim '(#\space #\return #\linefeed) line))))
    (if (zerop (length trimmed))
        (values nil nil)
        (let ((tokens (split-sequence:split-sequence #\space trimmed)))
          (values (string-upcase (first tokens))
                  (rest tokens))))))

(defun read-elements (stream expected-type creator)
  (let ((line (read-line stream nil nil)))
    (multiple-value-bind (type elements)
        (parse-response line)
      (assert (string= type expected-type))
      (loop for element in elements
            as parts = (split-sequence:split-sequence #\, element)
            collect (apply creator parts)))))

(defun read-blocks-state (stream)
  (read-elements stream "BLOCKS"
                 (lambda (x y) (cons (parse-integer x) (parse-integer y)))))

(defun read-players-state (stream)
  (read-elements stream "PLAYERS"
                 (lambda (index x y dx dy direction moving-p status)
                   (make-instance 'player
                                  :x (parse-integer x) :y (parse-integer y)
                                  :dx (parse-integer dx) :dy (parse-integer dy)
                                  :direction (intern direction :keyword)
                                  :moving-p (string/= moving-p "NIL")
                                  :status (intern status :keyword)))))

(defun read-bombs-state (stream)
  (read-elements stream "BOMBS"
                 (lambda (x y power step)
                   (make-instance 'bomb
                                  :x (parse-integer x)
                                  :y (parse-integer y)
                                  :power (parse-integer power)
                                  :explosion-step (if (string= step "NIL")
                                                      nil
                                                      (parse-integer step))))))

(defun parse-map (string width height)
  (let ((data (make-array (list height width) :initial-element nil))
        (index 0))
    (loop for j below height
          do (loop for i below width
                   do (setf (aref data j i)
                            (ecase (aref string index)
                              (#\. nil)
                              (#\X :wall)
                              (#\O :obstacle)))
                      (incf index)))
    (make-instance 'game-map :width width :height height :data data)))

(defun play-game (server-stream map)
  (pal:with-pal (:width +screen-width+ :height +screen-height+
                 :fullscreenp nil :fps 50)
    (load-images)
    (load-sounds)
    (let ((frame 0)
          (last-bomb-frame 0))
      (pal:event-loop ()
        (format server-stream "START~%")
        (pal:test-keys
          (:key-up (format server-stream "MOVE NORTH~%"))
          (:key-down (format server-stream "MOVE SOUTH~%"))
          (:key-left (format server-stream "MOVE WEST~%"))
          (:key-right (format server-stream "MOVE EAST~%"))
          (:key-space
           (when (> frame (1+ last-bomb-frame))
             (format server-stream "BOMB~%")
             (setf last-bomb-frame frame)))
          (:key-q
           (format server-stream "QUIT~%")
           (finish-output server-stream)
           (return-from play-game)))
        (format server-stream "GETSTATE~%")
        (finish-output server-stream)
        (let* ((blocks (read-blocks-state server-stream))
               (players (read-players-state server-stream))
               (bombs (read-bombs-state server-stream)))
          (when (and (member 0 bombs :key 'explosion-step) *explosion-sound*)
            (pal:play-sample *explosion-sound* :loops nil))
          (loop for bomb in bombs
                when (eql (explosion-step bomb) 0)
                do (loop for (x . y) in (bomb-kill-zone bomb map)
                         do (setf (aref (data map) y x) nil)))
          (draw-game map blocks players bombs))
        (incf frame)))))

(defun blaster (server-host server-port player-name)
  (format t "Connecting to server ~A:~A...~%" server-host server-port)
  (let* ((server-socket (usocket:socket-connect server-host server-port))
         (server-stream (usocket:socket-stream server-socket)))
    (format server-stream "PLAY ~A~%" player-name)
    (finish-output server-stream)
    (let ((line (read-line server-stream nil nil)))
      (multiple-value-bind (response args)
          (parse-response line)
        (cond ((null response)
               (format t "Connection error, stopping.~%"))
              ((string= response "ERROR")
               (format t "Cannot join game:~A~%" (subseq line 5)))
              ((string= response "OK")
               (format t "Joining game, map ~Ax~A~%" (first args) (second args))
               (let* ((map-width (parse-integer (first args)))
                      (map-height (parse-integer (second args)))
                      (map (parse-map (third args) map-width map-height)))
                 (play-game server-stream map))))))))

(in-package :cl-user)

(defun start ()
  (blaster-client:blaster "localhost" 4000 "matt"))
