(require :usocket)
(require :split-sequence)

(load "common")

(defpackage :blaster-server
  (:use :common-lisp :blaster-common)
  (:export #:start #:stop))

(in-package :blaster-server)

;;; Logging facility

(defparameter *log-file* "/tmp/gameserver.log")

(defun level-string (level)
  (ecase level
    (:error "error")
    (:warning "warning")
    (:info "info")
    (:debug "debug")))

(defun log-event (level format &rest args)
  (multiple-value-bind (second minute hour date month year day daylight-p zone)
      (decode-universal-time (get-universal-time))
    (declare (ignore day daylight-p zone))
    (let ((message (apply #'format nil format args)))
      (with-open-file (stream *log-file*
                       :direction :output
                       :if-does-not-exist :create
                       :if-exists :append)
        (format stream "~4,'0D-~2,'0D-~2,'0D ~2,'0D:~2,'0D:~2,'0D [~A] ~A~%"
                year month date hour minute second
                (level-string level)
                message)))))

;;; Game

(defvar *map*)

(defvar *max-player-count*)
(defvar *players*)

(defvar *bombs*)

(defun make-map (width height)
  (let ((data (make-array (list height width) :initial-element nil)))
    (loop for i below width
          do (setf (aref data 0 i) :wall
                   (aref data (1- height) i) :wall))
    (loop for j below height
          do (setf (aref data j 0) :wall
                   (aref data j (1- width)) :wall))
    (loop for j from 2 to (- height 3) by 2
          do (loop for i from 2 to (- width 3) by 2
                   do (setf (aref data j i) :wall)))
    (loop repeat 10
          do (setf (aref data
                         (+ (* (random (floor (- height 1) 2)) 2) 1)
                         (+ (* (random (floor (- width 1) 2)) 2) 1))
                   :obstacle))
    (make-instance 'game-map :width width :height height :data data)))

(defun map->string (map)
  (with-output-to-string (stream)
    (loop for j below (height map)
          do (loop for i below (width map)
                   do (write-char (ecase (aref (data map) j i)
                                    ((nil) #\.)
                                    (:wall #\X)
                                    (:obstacle #\O))
                                  stream)))))

(defun blocking-p (cell)
  (member cell '(:wall :obstacle)))

(defun spawn-player (player-name map)
  (let ((player (make-instance 'player
                               :name player-name
                               :bomb-count 3
                               :bomb-power 2
                               :dropped-bombs '()
                               :x (random (width map))
                               :y (random (height map))
                               :dx 0
                               :dy 0
                               :direction :south
                               :moving-p nil
                               :status :playing)))
    (push player *players*)
    player))

(defun kill-player (player)
  (setf *players* (remove player *players*)))

(defun move-player (player map)
  (let ((data (data map)))
    (case (direction player)
      (:north
       (setf (dx player) 0)
       (decf (dy player) +player-step+)
       (when (and (< (dy player) 0)
                  (blocking-p (aref data (1- (y player)) (x player))))
         (setf (dy player) 0))
       (when (< (dy player) (- +cell-half-size+))
         (decf (y player))
         (incf (dy player) +cell-size+)))
      (:south
       (setf (dx player) 0)
       (incf (dy player) +player-step+)
       (when (and (> (dy player) 0)
                  (blocking-p (aref data (1+ (y player)) (x player))))
         (setf (dy player) 0))
       (when (>= (dy player) +cell-half-size+)
         (incf (y player))
         (decf (dy player) +cell-size+)))
      (:west
       (setf (dy player) 0)
       (decf (dx player) +player-step+)
       (when (and (< (dx player) 0)
                  (blocking-p (aref data (y player) (1- (x player)))))
         (setf (dx player) 0))
       (when (< (dx player) (- +cell-half-size+))
         (decf (x player))
         (incf (dx player) +cell-size+)))
      (:east
       (setf (dy player) 0)
       (incf (dx player) +player-step+)
       (when (and (> (dx player) 0)
                  (blocking-p (aref data (y player) (1+ (x player)))))
         (setf (dx player) 0))
       (when (>= (dx player) +cell-half-size+)
         (incf (x player))
         (decf (dx player) +cell-size+))))))

(defun drop-bomb (player)
  (let ((bomb (make-instance 'bomb
                             :x (x player)
                             :y (y player)
                             :owner player
                             :power (bomb-power player)
                             :ticks 50
                             :explosion-step nil)))
    (push bomb *bombs*)
    (push bomb (dropped-bombs player))
    bomb))

(defun explode (bomb map)
  (loop for (x . y) in (bomb-kill-zone bomb map)
        do (setf (aref (data map) y x) nil)
        do (loop for player in *players*
                 when (and (= (x player) x)
                           (= (y player) y))
                 do (setf (status player) :dead))))

(defun update-game ()
  (loop for bomb in *bombs*
        do (if (> (ticks bomb) 0)
               (decf (ticks bomb))
               (cond ((null (explosion-step bomb))
                      (setf (explosion-step bomb) 0)
                      (explode bomb *map*)
                      (let ((owner (owner bomb)))
                        (setf (dropped-bombs owner)
                              (remove bomb (dropped-bombs owner)))))
                     ((< (explosion-step bomb) 5)
                      (incf (explosion-step bomb)))
                     (t
                      (setf *bombs* (remove bomb *bombs*))))))
  (sleep 0.1))

;;; Server

(defvar *server-socket*)

(defpackage :blaster-commands)

(defun parse-command (line)
  (let ((trimmed (and line (string-trim '(#\space #\return #\linefeed) line))))
    (if (zerop (length trimmed))
        (values nil nil)
        (let ((tokens (split-sequence:split-sequence #\space trimmed)))
          (values (intern (string-upcase (first tokens)) 'blaster-commands)
                  (rest tokens))))))

(defun write-state (stream)
  (format stream "BLOCKS")
  (loop for j below (height *map*)
        do (loop for i below (width *map*)
                 when (eql (aref (data *map*) j i) 'block)
                 do (format stream " ~A,~A" i j)))
  (format stream "~%PLAYERS")
  (loop for player in *players*
        for index from 0
        do (format stream " ~A,~A,~A,~A,~A,~A,~A,~A"
                   index
                   (x player) (y player)
                   (dx player) (dy player)
                   (symbol-name (direction player))
                   (moving-p player)
                   (symbol-name (status player))))
  (format stream "~%BOMBS")
  (loop for bomb in *bombs*
        do (format stream " ~A,~A,~A,~A"
                   (x bomb) (y bomb) (power bomb) (explosion-step bomb)))
  (format stream "~%"))

(defun handle-client (socket)
  (log-event :info "New connection: ~A" socket)
  (loop with stream = (usocket:socket-stream socket)
        with player = nil
        as line = (read-line stream nil nil)
        as input = (and line (string-trim '(#\space #\return #\linefeed) line))
        do (multiple-value-bind (command args)
               (parse-command line)
             (case command
               ((nil blaster-commands::quit)
                (when player
                  (kill-player player))
                (return))
               (blaster-commands::play
                (cond ((= (length *players*) *max-player-count*)
                       (format stream "ERROR too many players~%")
                       (return))
                      (t
                       (setf player (spawn-player (first args) *map*))
                       (format stream "OK ~A ~A ~A~%"
                               (width *map*) (height *map*)
                               (map->string *map*)))))
               (blaster-commands::start
                (setf (moving-p player) nil))
               (blaster-commands::move
                (when (eq (status player) :playing)
                  (setf (direction player) (intern (first args) :keyword))
                  (move-player player *map*)
                  (setf (moving-p player) t)))
               (blaster-commands::bomb
                (if (or (= (length (dropped-bombs player)) (bomb-count player))
                        (eq (status player) :dead))
                    (log-event :warning "Illegal attempt to drop a bomb: ~A"
                               socket)
                    (drop-bomb player)))
               (blaster-commands::getstate
                (write-state stream))))
           (finish-output stream))
  (usocket:socket-close socket)
  (log-event :info "Done processing ~A" socket))

(defun server-listen ()
  (let ((socket (usocket:socket-accept *server-socket*)))
    (sb-thread:make-thread (lambda () (handle-client socket))
                           :name "Game Server [client]")))

;;; Start / stop

(defvar *listen-thread*)
(defvar *update-thread*)

(defun start (port map-width map-height max-player-count)
  (setf *map* (make-map map-width map-height))
  (setf *max-player-count* max-player-count
        *players* '()
        *bombs* '())
  (setf *server-socket* (usocket:socket-listen "localhost" port
                                               :reuseaddress t))
  (setf *listen-thread* (sb-thread:make-thread (lambda ()
                                                 (loop (server-listen)))
                                               :name "Game Server [listen]"))
  (setf *update-thread* (sb-thread:make-thread (lambda ()
                                                 (loop (update-game)))
                                               :name "Game Server [update]"))
  (log-event :info "Server started on port ~A" port)
  t)

(defun stop ()
  (usocket:socket-close *server-socket*)
  (sb-thread:terminate-thread *listen-thread*)
  (sb-thread:terminate-thread *update-thread*)
  (log-event :info "Server stopped")
  t)
